program DNIeReader;

{$APPTYPE CONSOLE}

uses
  Windows,
  SysUtils,
  Classes,
  WinSCard in 'WinSCard.pas';


function Min(i,j: Integer): Integer;
begin
  if i < j then
    Result:= i
  else
    Result:= j;
end;
 
// Muestra el contenido del buffer en hexadecimal y como texto
procedure WriteHex(Buffer: PAnsiChar;  Count: Integer);
var
  i,j: Integer;
begin
  j:= 0;
  while Count > 0 do
  begin
    Write(IntToHex(j,8) + ':' + #32#32);
    for i:= 0 to Min(Count,8) - 1 do
      Write(IntToHex(Byte(Buffer[i]),2) + #32);
    Write(#32);
    for i:= 8 to Min(Count,16) - 1 do
      Write(IntToHex(Byte(Buffer[i]),2) + #32);
    for i:= Min(Count,16) to 15 do
      Write(#32#32#32);
    Write(#32 + '|');
    for i:= 0 to Min(Count,16) - 1 do
    if Char(Buffer[i]) in ['A'..'Z','a'..'z','0'..'9'] then
      Write(Buffer[i])
    else
      Write('.');
    Writeln('|');
    Dec(Count,16);
    inc(Buffer,16);
    inc(j,16);
  end;
end;

function isDNIe(hCard: SCARDHANDLE): Boolean;
var
  bAtr: PByteArray;
  cbAtrLen: DWORD;
  lReturn: Longint;
begin
  bAtr:= nil;
  cbAtrLen:= 0;
  lReturn:= SCardStatus(hCard, nil,nil,nil,nil,LPBYTE(bAtr),@cbAtrLen);
  // http://www.dnielectronico.es/PDFs/ATR1.pdf
  if (lReturn <> SCARD_S_SUCCESS) or (cbAtrLen<>20) then
  begin
    Result:= FALSE;
    Exit;
  end;
  GetMem(bAtr,cbAtrLen);
  try
    lReturn:= SCardStatus(hCard, nil,nil,nil,nil,LPBYTE(bAtr),@cbAtrLen);
    if lReturn <> SCARD_S_SUCCESS then
    begin
      Result:= FALSE;
      Exit;
    end;
    // http://www.dnielectronico.es/PDFs/ATR1.pdf
    Result:=
      (bAtr[00]=$3B) and (bAtr[01]=$7F) and (bAtr[03]=$00) and (bAtr[04]=$00) and
      (bAtr[05]=$00) and (bAtr[06]=$6A) and (bAtr[07]=$44) and (bAtr[08]=$4E) and
      (bAtr[09]=$49) and (bAtr[10]=$65) and (bAtr[18]=$90) and (bAtr[19]=$00);
  finally
    FreeMem(bAtr);
  end;
end;

// Envia un comando a la tarjeta (iso_7816-4v2_0.pdf)
function SendCmd(Card: SCARDHANDLE; Cmd: AnsiString; Buffer: PByte; var l: DWORD): Boolean;
begin
  Result:= FALSE;
  if SCardTransmit(Card, @SCARD_PCI_T0,PAnsiChar(Cmd),Length(Cmd), nil,
    LPBYTE(Buffer), @l) = SCARD_S_SUCCESS then
    if l >= 2 then
      Result:= (PByteArray(Buffer)[l-2] = $61) or
        ((PByteArray(Buffer)[l-2] = $90) and (PByteArray(Buffer)[l-1] = $00));
end;

// http://www.dnie.es/PDFs/obtiene_CDF.pdf
function SCardReadFile(Card: SCARDHANDLE; Path: AnsiString; Stream: TStream): Boolean;
var
  Buffer: PByte;
  l: DWORD;
  Offset: Word;
  Size: Word;
begin
  Result:= FALSE;
  if Odd(Length(Path)) then
    Exit;
  GetMem(Buffer,258);
  try
    l:= 258;
    // SELECT directorio raiz
    if not SendCmd(Card,
      #$00#$a4#$04#$00#$0b#$4D#$61#$73#$74#$65#$72#$2E#$46#$69#$6C#$65,
      Buffer, l) then Exit;
    // Seguimos la ruta del fichero
    while Length(Path) > 0 do
    begin
      l:= 258;
      // SELECT
      if not SendCmd(Card,#$00#$A4#$00#$00#$02 + Copy(Path,1,2),
        Buffer, l) then Exit;
      delete(Path,1,2);
    end;
    l:= 258;
    // GET RESPONSE
    if not SendCmd(Card,#$00#$C0#$00#$00#$0E,Buffer, l) then
      Exit;
    if (l <> $10) or (PByteArray(buffer)[0] <> $6F) then
      Exit;
    // Obtenemos el tamao
    Size:= (PByteArray(buffer)[7] * $100) + PByteArray(buffer)[8];
    Offset:= 0;
    while Offset < Size do
    begin
      l:= 258;
      // Leemos el fichero en bloques de hasta 255 bytes
      if not SendCmd(Card,#$00#$B0 + AnsiChar(Hi(Offset)) + AnsiChar(Lo(Offset))
        + Char(Min(Size-Offset,$FF)), Buffer, l) then Exit;
      if l > 2 then
        Stream.Write(Buffer^,l-2);
      inc(Offset,$FF);
    end;
    Result:= TRUE;
  finally
    FreeMem(Buffer);
  end;
end;

var
  hContext: SCARDCONTEXT;
  pmszReaders: PAnsiChar;
  pReader: PAnsiChar;
  cchReaders: DWORD;
  lReturn: Longint;
  Card: SCARDHANDLE;
  ActiveProtocol: DWORD;
  Stream: TMemoryStream;
begin
  lReturn:= SCardEstablishContext(SCARD_SCOPE_SYSTEM,nil,nil,@hContext);
  if lReturn = SCARD_S_SUCCESS then
  try
    cchReaders:= SCARD_AUTOALLOCATE;
    // Obtenemos una lista con todos los lectores
    lReturn:= SCardListReaders(hContext,nil,@pmszReaders,@cchReaders);
    if lReturn = SCARD_S_SUCCESS then
    begin
      pReader:= pmszReaders;
      // Intentamos conectarnos a cada uno de los lectores de la lista
      while (byte(pReader^) <> 0) do
      begin
        lReturn:= SCardConnect(hContext, pReader, SCARD_SHARE_SHARED,
          SCARD_PROTOCOL_T0, @Card, @ActiveProtocol);
        if lReturn = SCARD_S_SUCCESS then
        begin
          // Comprobamos si la tarjeta insertada es un DNIe
          if isDNIe(Card) then
          begin
            Writeln('DNIe Encontrado');
            Writeln;
            Stream:= TMemoryStream.Create;
            try
              // Leemos el IDESP con la ruta 0006
              if SCardReadFile(Card,#$00#$06,Stream) then
              begin
                Writeln('IDESP (0006):');
                WriteHex(Stream.Memory,Stream.Size);
                Writeln;
                Stream.SaveToFile(IncludeTrailingPathDelimiter(
                  ExtractFilePath(ParamStr(0))) + 'dump\IDESP_0006.bin');
              end;
              Stream.Clear;
              // Leemos la version con la ruta 2F03
              if SCardReadFile(Card,#$2F#$03,Stream) then
              begin
                Writeln('Version (2F03):');
                WriteHex(Stream.Memory,Stream.Size);
                Writeln;
                Stream.SaveToFile(IncludeTrailingPathDelimiter(
                  ExtractFilePath(ParamStr(0))) + 'dump\Version_2F03.bin');
              end;
              Stream.Clear;
              // Leemos el CDF con la ruta 50156004
              if SCardReadFile(Card,#$50#$15#$60#$04,Stream) then
              begin
                Writeln('CDF (50156004):');
                WriteHex(Stream.Memory,Stream.Size);
                Writeln;
                Stream.SaveToFile(IncludeTrailingPathDelimiter(
                  ExtractFilePath(ParamStr(0))) + 'dump\CDF_50156004.bin');
              end;
            finally
              Stream.Free;
            end;
            SCardDisconnect(Card,SCARD_RESET_CARD);
            break;
          end;
          SCardDisconnect(Card,SCARD_RESET_CARD);
        end;
        pReader:= pReader + StrLen(pReader) + 1;
      end;
      SCardFreeMemory(hContext,pmszReaders);
    end;
  finally
    SCardReleaseContext(hContext);
  end;
  Writeln('Adios!');
end.
